from pathlib import Path


UNKNOW_VERSION = "unknown (git checkout)"


def string() -> str:
    try:
        fp = Path(__file__).parent / "VERSION"
        if not fp.is_file():
            return UNKNOW_VERSION
        with open(fp, "r", encoding="utf-8") as fh:
            version = fh.read().strip()
            if version:
                return version
    except:
        pass
    return UNKNOW_VERSION



"""Top-level package for Async SQL CRUD."""

__author__ = """Lucas Lucyk"""
__email__ = "llucyk@gmail.com"
# __version__ = "0.0.1"

from .crud import CRUDBase
from .exceptions import CreateException, NotFoundException
from .models import ModelBase


__all__ = ["CRUDBase", "CreateException", "NotFoundException", "ModelBase"]



from datetime import datetime
from sqlalchemy import Column, DateTime
from sqlalchemy.orm import declarative_mixin


@declarative_mixin
class Timestamp:
    created_at = Column(DateTime, default=datetime.utcnow, nullable=False)
    updated_at = Column(DateTime, default=datetime.utcnow, nullable=False)


from datetime import datetime
from typing import (
    Any,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)
from uuid import UUID
from pydantic import BaseModel
from sqlalchemy import select
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession
from ..models.base import ModelBase
from ..exceptions.crud import NotFoundException, CreateException


ModelType = TypeVar("ModelType", bound=ModelBase)
CreateSchemaType = TypeVar("CreateSchemaType", bound=BaseModel)
UpdateSchemaType = TypeVar("UpdateSchemaType", bound=BaseModel)


class CRUDBase(Generic[ModelType, CreateSchemaType, UpdateSchemaType]):
    def __init__(self, model: Type[ModelType]) -> None:
        """CRUD object with default methods to Create, Read, Update, & Delete

        Args:
            model (Type[ModelType]): A SQLAlchemy model class
        """
        self.model = model

    async def get(self, db: AsyncSession, uid: UUID) -> Optional[ModelType]:
        """Get row from model where id == model.id

        Args:
            db (AsyncSession): Async db session
            uid (UUID): UUID to filter

        Returns:
            Optional[ModelType]: ModelType instance or None if id not exists
        """

        res = await db.execute(select(self.model).where(self.model.uid == uid))
        return res.scalar()

    async def get_or_raise(
        self, db: AsyncSession, uid: UUID
    ) -> Optional[ModelType]:
        """Try to dgt row from model where id == model.id

        Args:
            db (AsyncSession): Async db session
            uid (UUID): UUID to filter

        Raises:
            NotFoundException: If item does not exist

        Returns:
            Optional[ModelType]: ModelType instance
        """

        # try get item
        obj = await self.get(db=db, uid=uid)

        if not obj:
            raise NotFoundException(f"{self.model.__name__} not found")

        return obj

    async def list(
        self, db: AsyncSession, *, skip: int = 0, limit: int = 100
    ) -> List[ModelType]:
        """Get multi items from database without filter criteria

        Args:
            db (AsyncSession): Async db session
            skip (int, optional): Optional Offset. Defaults to 0.
            limit (int, optional): Optional limit. Defaults to 100.

        Returns:
            List[ModelType]: Matching results list
        """

        results = await db.execute(select(self.model).offset(skip).limit(limit))
        return results.scalars().all()

    async def filter(
        self,
        db: AsyncSession,
        whereclause: Any,
        *,
        skip: int = 0,
        limit: int = 100,
    ) -> List[ModelType]:
        """Get items from database using `filters` to filter

        Args:
            db (AsyncSession): Async db session
            whereclause (Any): Whereclause to filter.
            skip (int, optional): Optional Offset. Defaults to 0.
            limit (int, optional): Optional limit. Defaults to 100.

        Returns:
            List[ModelType]:
                Instance or list of intance of matching items.
        """

        # try to get
        result = await db.execute(
            select(self.model).where(whereclause).offset(skip).limit(limit)
        )

        return result.scalars().all()

    async def find(
        self, db: AsyncSession, *, skip: int = 0, limit: int = 100, **kwargs
    ) -> List[ModelType]:
        """Find elements with kwargs

        Args:
            db (AsyncSession): Async db session
            skip (int, optional): Optional Offset. Defaults to 0.
            limit (int, optional): Optional limit. Defaults to 100.

        Returns:
            List[ModelType]: list of results
        """

        result = await db.execute(
            select(self.model).filter_by(**kwargs).offset(skip).limit(limit)
        )

        return result.scalars().all()

    async def find_one(self, db: AsyncSession, **kwargs) -> ModelType:
        """Find an element with kwargs

        Args:
            db (AsyncSession): Async db session

        Returns:
            ModelType: First result object
        """

        result = await db.execute(
            select(self.model).filter_by(**kwargs).limit(1)
        )

        return result.scalar()

    @staticmethod
    def _set_updated_at(element: ModelType) -> None:
        """Set `updated_at` property to current time if property exists

        Args:
            obj (ModelType): Object instance
        """

        if hasattr(element, "updated_at"):
            setattr(element, "updated_at", datetime.utcnow())

    async def save(self, db: AsyncSession, element: ModelType) -> ModelType:
        """Save an object into database

        Args:
            db (AsyncSession): Async db session
            element (ModelType): ModelType instance to save

        Returns:
            ModelType: Saved object instance
        """

        # set updated_at
        self._set_updated_at(element)

        # add, commit and refresh
        db.add(element)
        await db.commit()
        await db.refresh(element)

        # return instance
        return element

    async def save_all(
        self, db: AsyncSession, elements: Iterable[ModelType]
    ) -> Iterable[ModelType]:
        """Save an iterable of elements into database

        Args:
            db (AsyncSession): Async db session
            elements (Iterable[ModelType]): Iterable of ModelType instance to save

        Returns:
            Iterable[ModelType]: Iterable of ModelType instance
        """

        # set updated_at
        _ = list(map(self._set_updated_at, elements))

        # add, commit and refresh
        db.add_all(elements)
        await db.commit()
        for element in elements:
            await db.refresh(element)

        # return instances
        return elements

    async def create(
        self, db: AsyncSession, element: CreateSchemaType
    ) -> ModelType:
        """Create an item into database

        Args:
            db (AsyncSession): Async db session
            element (CreateSchemaType): Schema to create

        Raises:
            CreateException: if item already exists or unexpected error

        Returns:
            ModelType: Instance of created object
        """

        try:
            # try json encode
            db_obj = self.model(**element.model_dump(mode="json"))
            return await self.save(db=db, element=db_obj)

        except IntegrityError:
            raise CreateException(f"{self.model.__name__} already exists.")

    async def bulk_create(
        self, db: AsyncSession, elements: Iterable[CreateSchemaType]
    ) -> Iterable[ModelType]:
        """Create items into database

        Args:
            db (AsyncSession): Async db session
            data (Iterable[CreateSchemaType]): Iterable of Schema to create

        Raises:
            CreateException: if an item already exists or unexpected error

        Returns:
            Iterable[ModelType]: Iterable of Instance of created object
        """

        try:
            # try json encode
            db_objs = [
                self.model(**d.model_dump(mode="python")) for d in elements
            ]
            return await self.save_all(db=db, elements=db_objs)

        except IntegrityError:
            raise CreateException(f"{self.model.__name__} already exists.")

    async def update(
        self,
        db: AsyncSession,
        *,
        obj: ModelType,
        data: Union[UpdateSchemaType, Dict[str, Any]],
    ) -> ModelType:
        """Update a database item with an update schema

        Args:
            db (AsyncSession): Async db session
            obj (ModelType): Item to update
            data (Union[UpdateSchemaType, Dict[str, Any]]):
                New partial or full data for database item

        Returns:
            ModelType: Instance of updated object
        """

        # obj to dict
        if isinstance(data, dict):
            update_data = data
        else:
            update_data = data.model_dump(exclude_unset=True)

        # update obj with each field of obj
        for field, value in update_data.items():
            setattr(obj, field, value)

        return await self.save(db=db, element=obj)

    async def delete(self, db: AsyncSession, uid: UUID) -> ModelType:
        """Delete an item from database

        Args:
            db (AsyncSession): Async db session
            id (int): Id of model to delete

        Returns:
            ModelType: Deleted object instance
        """

        obj = await self.get_or_raise(db=db, uid=uid)

        await db.delete(obj)
        await db.commit()

        return obj



from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.asyncio import (
    AsyncSession,
    AsyncEngine,
    create_async_engine,
    async_sessionmaker
)
from .config import DB_URI, DB_URI_ASYNC


# async sqlite session
async_engine: AsyncEngine = create_async_engine(
    DB_URI_ASYNC,
    future=True,
    connect_args={"check_same_thread": False}
)

AsyncSessionLocal: AsyncSession = async_sessionmaker(
    bind=async_engine,
    class_=AsyncSession,
    expire_on_commit=False
)

# sqlite session
engine = create_engine(
    DB_URI, future=True, connect_args={"check_same_thread": False}
)
SessionLocal = sessionmaker(expire_on_commit=False, bind=engine)




from typing import Optional
from uuid import UUID
from pydantic import BaseModel


class SampleBase(BaseModel):
    id: Optional[int] = None
    uid: Optional[UUID] = None
    token: Optional[str] = None
    describe: Optional[str] = None
    email: Optional[str] = None


class SampleCreate(SampleBase):
    ...


class SampleUpdate(SampleBase):
    ...


class SampleInDB(SampleBase):

    class Config:
        from_attributes: True


class Sample(SampleInDB):
    ...



from typing import Any
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.ext.declarative import declared_attr


class ModelBase(DeclarativeBase):
    id: Any
    uid: Any
    __name__: str

    # Generate __tablename__ automatically
    @declared_attr
    def __tablename__(cls) -> str:
        return cls.__name__.lower()



from ..crud import CRUDBase
from .models import Sample as SampleModel
from .schemas import SampleCreate, SampleUpdate



class CRUDSample(CRUDBase[SampleModel, SampleCreate, SampleUpdate]):
    ...


samples = CRUDSample(SampleModel)


from uuid import uuid1, UUID
from sqlalchemy.orm import Mapped, mapped_column
from secrets import token_urlsafe
from ..models import ModelBase, Timestamp


class Sample(Timestamp, ModelBase):
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    uid: Mapped[UUID] = mapped_column(index=True, unique=True, default=uuid1)
    token: Mapped[str] = mapped_column(default=token_urlsafe)
    describe: Mapped[str] = mapped_column(nullable=True)
    email: Mapped[str] = mapped_column(nullable=True, unique=True)


import os
from pathlib import Path
from pydoc import describe
import unittest
from uuid import uuid1
from sqlalchemy.sql import text

from ..models import ModelBase
from ..exceptions.crud import NotFoundException, CreateException
from .schemas import SampleCreate, SampleUpdate
from .models import Sample
from .session import AsyncSessionLocal, engine
from .crud import samples
from .config import DB_NAME


class TestCRUD(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.db_path = Path(__file__).absolute().parent.parent.parent / DB_NAME
        if cls.db_path.is_file():
            os.remove(cls.db_path)
        ModelBase.metadata.create_all(bind=engine)

    @classmethod
    def tearDownClass(cls) -> None:
        ModelBase.metadata.drop_all(bind=engine)
        if cls.db_path.is_file():
            os.remove(cls.db_path)
        super().tearDownClass()

    async def test_create(self) -> None:
        async with AsyncSessionLocal() as db:
            email = "fake-create@fakedomain.com"
            sample_create = SampleCreate(email=email, token="fake-token-1")
            sample_obj = await samples.create(db=db, element=sample_create)

            self.assertIsInstance(sample_obj, Sample)
            self.assertEqual(sample_obj.token, sample_create.token)

            # create with raise
            with self.assertRaises(CreateException):
                error = SampleCreate(email=email)
                await samples.create(db=db, element=error)

    async def test_bulk_create(self) -> None:
        async with AsyncSessionLocal() as db:
            email = "fake-bulk-create@fakedomain.com"
            sample_2 = SampleCreate(
                email=email, token="fake-token-2", describe="text"
            )
            sample_3 = SampleCreate(token="fake-token-3", describe="text")

            sample_2_db, sample_3_db = await samples.bulk_create(
                db=db, elements=(sample_2, sample_3)
            )
            self.assertIsInstance(sample_2_db, Sample)
            self.assertIsInstance(sample_3_db, Sample)
            self.assertEqual(sample_2_db.token, sample_2.token)
            self.assertEqual(sample_3_db.token, sample_3.token)

            # create with raise
            with self.assertRaises(CreateException):
                s = SampleCreate(describe="text-ok")
                error = SampleCreate(email=email)
                await samples.bulk_create(db=db, elements=(s, error))

    async def test_list(self) -> None:
        async with AsyncSessionLocal() as db:
            all_ = await samples.list(db=db)
            self.assertGreater(len(all_), 0, "DB items were expected")
            for obj in all_:
                self.assertIsInstance(obj, Sample)

            limited = await samples.list(db=db, limit=1)
            self.assertEqual(len(limited), 1)

            skipped = await samples.list(db=db, skip=1)
            self.assertLess(len(skipped), len(all_))

    async def test_filter(self) -> None:
        async with AsyncSessionLocal() as db:
            single = await samples.filter(
                db=db, whereclause=samples.model.token == "fake-token-1"
            )
            self.assertEqual(len(single), 1, "One single item was expected")
            self.assertIsInstance(single[0], Sample)

            wc = text("token LIKE '%fake%'")
            multi = await samples.filter(db=db, whereclause=wc)
            self.assertGreater(len(multi), 1)
            for obj in multi:
                self.assertIsInstance(obj, Sample)

            skipped_limited = await samples.filter(
                db=db, whereclause=wc, skip=1, limit=1
            )
            self.assertEqual(len(skipped_limited), 1)

    async def test_find(self) -> None:
        async with AsyncSessionLocal() as db:
            single = await samples.find(db=db, token="fake-token-1")
            self.assertEqual(len(single), 1, "One single item was expected")
            self.assertIsInstance(single[0], Sample)

            multi = await samples.find(db=db, describe="text")
            self.assertGreater(len(multi), 1)
            for obj in multi:
                self.assertIsInstance(obj, Sample)

            limited = await samples.find(db=db, describe="text", limit=1)
            self.assertEqual(len(limited), 1, "One single item was expected")

            skipped = await samples.find(db=db, describe="text", skip=1)
            self.assertEqual(len(skipped), 1, "One single item was expected")

            not_found = await samples.find(db=db, describe="not-found")
            self.assertEqual(len(not_found), 0)

            skip_all = await samples.find(db=db, describe="text", skip=5)
            self.assertEqual(len(skip_all), 0)

    async def test_find_one(self) -> None:
        async with AsyncSessionLocal() as db:
            # unique with some data
            unique = await samples.find_one(db=db, describe=None)
            self.assertIsInstance(unique, Sample)

            # first of multi
            multi = await samples.find_one(db=db, describe="text")
            self.assertIsInstance(multi, Sample)

            # first of db
            empty = await samples.find_one(db=db)
            self.assertIsInstance(empty, Sample)

            # not found data
            not_found = await samples.find_one(db=db, describe="not-found")
            self.assertIsNone(not_found)

    async def test_get(self) -> None:
        async with AsyncSessionLocal() as db:
            # get all samples
            all_samples = await samples.list(db=db)

            sample = await samples.get(db=db, uid=all_samples[0].uid)
            self.assertIsInstance(sample, Sample)
            self.assertEqual(sample.uid, all_samples[0].uid)

            # not found
            empty = await samples.get(db=db, uid=uuid1())
            self.assertIsNone(empty)

    async def test_get_or_raise(self) -> None:
        async with AsyncSessionLocal() as db:
            # get all samples
            all_samples = await samples.list(db=db, skip=1, limit=1)

            # with raise
            sample = await samples.get_or_raise(db=db, uid=all_samples[0].uid)
            self.assertIsInstance(sample, Sample)
            self.assertEqual(sample.id, all_samples[0].id)

            # not found with raise
            with self.assertRaises(NotFoundException):
                await samples.get_or_raise(db=db, uid=uuid1())

    async def test_save(self) -> None:
        async with AsyncSessionLocal() as db:
            # get element
            obj = await samples.find_one(db=db, describe=None)
            updated_at = obj.updated_at
            uid = obj.uid
            id_ = obj.id

            # update and save
            obj.describe = "test-save"
            obj = await samples.save(db=db, element=obj)
            self.assertEqual(obj.describe, "test-save")

            self.assertNotEqual(updated_at, obj.updated_at)
            self.assertNotEqual(obj.created_at, obj.updated_at)
            self.assertEqual(uid, obj.uid)
            self.assertEqual(id_, obj.id)

    async def test_save_all(self) -> None:
        async with AsyncSessionLocal() as db:
            # get element
            objs = await samples.find(db=db, describe="text")
            updateds = [obj.updated_at for obj in objs]
            uids = [obj.uid for obj in objs]
            ids = [obj.id for obj in objs]

            # update and save
            for obj in objs:
                obj.describe = "test-save-all"

            # save
            objs = await samples.save_all(db=db, elements=objs)

            # check updated
            for obj in objs:
                self.assertEqual(obj.describe, "test-save-all")

            updateds_ = [obj.updated_at for obj in objs]
            uids_ = [obj.uid for obj in objs]
            ids_ = [obj.id for obj in objs]

            self.assertNotEqual(updateds, updateds_)
            self.assertEqual(uids, uids_)
            self.assertEqual(ids, ids_)

    async def test_update(self) -> None:
        async with AsyncSessionLocal() as db:
            sample = SampleCreate(describe="test-update")
            obj = await samples.create(db=db, element=sample)
            updated_ = obj.updated_at
            uid = obj.uid

            data = SampleUpdate(describe="test-update")
            obj = await samples.update(db=db, obj=obj, data=data)

            self.assertEqual(obj.describe, "test-update")
            self.assertEqual(obj.uid, uid)
            self.assertNotEqual(obj.updated_at, updated_)

            obj = await samples.update(
                db=db, obj=obj, data={"describe": "test-update-2"}
            )
            self.assertEqual(obj.describe, "test-update-2")

    async def test_delete(self) -> None:
        async with AsyncSessionLocal() as db:
            sample = SampleCreate(describe="for-test-delete")
            obj = await samples.create(db=db, element=sample)

            deleted = await samples.delete(db=db, uid=obj.uid)
            self.assertEqual(obj, deleted)

            not_found = await samples.find_one(
                db=db, describe="for-test-delete"
            )
            self.assertIsNone(not_found)



from setuptools import setup, find_packages
from pathlib import Path
import subprocess


def get_version_from_git_tag() -> str:
    version = (
        subprocess.run(["git", "describe", "--tags"], stdout=subprocess.PIPE)
        .stdout.decode("utf-8")
        .strip()
    )

    if "-" in version:
        # when not on tag, git describe outputs: "1.3.3-22-gdf81228"
        # pip has gotten strict with version numbers
        # so change it to: "1.3.3+22.git.gdf81228"
        # See: https://peps.python.org/pep-0440/#local-version-segments
        v, i, s = version.split("-")
        version = v + "+" + i + ".git." + s

    assert "-" not in version
    assert "." in version

    return version


sa_modelcrud_version: str = get_version_from_git_tag()

# write version on VERSION file
assert Path("sa_modelcrud/version.py").is_file()
with open("sa_modelcrud/VERSION", "w", encoding="utf-8") as fh:
    fh.write("%s\n" % sa_modelcrud_version)


# get description from README
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name="sa_modelcrud",
    version=sa_modelcrud_version,
    description="Base CRUD manager to manage databases with asynchronous SQLAlchemy sessions",
    packages=find_packages(),
    package_data={"sa_modelcrud": ["VERSION"]},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.8",
    install_requires=[
        "pydantic>=2.4.2",
        "sqlalchemy>=2.0.22"
    ],
    url="https://github.com/lucaslucyk/sa-model-crud",
    author="Lucas Lucyk",
    author_email="lucaslucyk@gmail.com",
)
